#ifndef _ARGPARSER_H_INCLUDED_
#define _ARGPARSER_H_INCLUDED_

#include <algorithm>
#include <cstdlib>
#include <cctype>
#include <map>
#include <iostream>
#include <string>
#include <stdexcept>
#include <vector>

using namespace std;

/**
 * A generic parser for command line options and arguments.  After
 * instantiation, the caller repeatly calls addSpec to supply the full
 * names of all options, and whether any one of them can have an
 * argument.  The caller then calls the parse method to pass in argc
 * and argv.
 *
 * To parse the command line inputs this parser sequentially goes
 * through each element in argv, skips the first one, which is the
 * program name.  If an element starts with a '-', it is assumed to be
 * an option, whose name is the substring of the element with leading
 * '-'s removed.  The name is then checked againest the option specs.
 * If it is the prefix or full match of an option name in the specs, a
 * match is found.  If canHasArg of the matching spec is true, and the
 * next element in argv does not start with a '-', then that element
 * is considered the argument or value of the option, otherwise the
 * value of the option is set to a zero length string.  The name value
 * pair is then stored into the internal option map.
 *
 * This parser supports parsing multiple command lines with the same
 * set of specs.  The parse method can be invoked multiple times.
 * Each time it clears results from the previous invocation before
 * parsing the given command line.
 *
 * All command line inputs after the last recongnizable option and its
 * argument are considered as arguments and returned to the caller
 * when getArgs method is called.
 *
 */
class ArgParser {

  map<string, bool> optSpecs;
  map<string, string> optMap; // key: opName, value: opValue
  vector<string> opts; // options appeared in argv
  vector<string> args; // tokens do not start with '-' and not follow an option

public:
  ArgParser() {}
  void addSpec(string fullOptName, bool canHasArg) throw (invalid_argument);
  void parse(int argc, char** argv) throw (invalid_argument);
  // return true if opt present in optMap
  bool hasOption(string prefix) throw (invalid_argument);
  string getString(string prefix) throw (invalid_argument);
  string getString(string prefix, string defaultVal) throw (invalid_argument);
  string getNonEmptyString(string prefix) throw (invalid_argument);
  bool getBool(string prefix) throw (invalid_argument);
  bool getBool(string prefix, bool defaultVal) throw (invalid_argument);
  int getInt(string prefix) throw (invalid_argument);
  int getInt(string prefix, int defaultVal) throw (invalid_argument);
  float getFloat(string prefix) throw (invalid_argument);
  float getFloat(string prefix, float defaultVal) throw (invalid_argument);
  double getDouble(string prefix) throw (invalid_argument);
  double getDouble(string prefix, double defaultVal) throw (invalid_argument);
  vector<string> const& getArgs() { return args; }

  // methods likely to be useful for debugging purpose, don't need to be fast
  int getNumSpecs() { return optSpecs.size(); }
  int getNumOpts() { return optMap.size(); }
  map<string, bool> const& getSpecs() { return optSpecs; }
  map<string, string> const& getOpts() { return optMap; }

private:
  // return zero length string if no spec with the prefix can be found
  string getOptName(string prefix) throw (invalid_argument);
};

inline void ArgParser::addSpec(string fullOptName, bool canHasArg)
  throw (invalid_argument)
{
  if (fullOptName[0] == '-') {
    throw invalid_argument("spec name cannot start with '-': " + fullOptName);
  } else if (optSpecs.find(fullOptName) != optSpecs.end()) {
    throw invalid_argument("spec for option \""+ fullOptName + 
                           "\" has been previously added alreay");
  }
  optSpecs.insert(make_pair(fullOptName, canHasArg));
}

inline void ArgParser::parse(int argc, char** argv) throw (invalid_argument) {
  // clear option map, incase multiple command lines need to be parsed
  optMap.clear();

  // collect options
  int i = 1; // argv[0] is the program name
  for (; i < argc; i++) {
    if (argv[i][0] == '-') {
      int j = 1;
      while (argv[i][j] == '-') {
	j++;
      }
      string prefix = &argv[i][j]; // skip the leading '-'s
      string optName = getOptName(prefix);
      string opValue = "";
      if (optSpecs[optName] && i + 1 < argc && argv[i + 1][0] != '-') {
	opValue = argv[++i];
      }
      if (optMap.find(optName) != optMap.end()) {
	throw invalid_argument("option \"" + getString(optName) + 
			       "\" appears more than once on command line");
      }
      optMap[optName] = opValue;
    } else {
      break; // done with all options
    }
  }

  // then collect arguments
  for (; i < argc; i++) {
    args.push_back(argv[i]);
  }
}

inline bool ArgParser::hasOption(string prefix) throw (invalid_argument) {
  string optName = getOptName(prefix);
  if (optMap.find(optName) != optMap.end()) {
    return true;
  } else {
    return false;
  }
}

inline string ArgParser::getString(string prefix) throw (invalid_argument) {
  string optName = getOptName(prefix);
  if (!hasOption(optName)) {
    throw invalid_argument("option \"" + optName + "\" cannot be found");
  }
  return optMap[optName];
}

inline string ArgParser::getString(string prefix, string defaultVal)
  throw (invalid_argument)
{
  string optName = getOptName(prefix);
  if (!hasOption(optName)) {
    return defaultVal;
  }
  return optMap[optName];
}

inline string ArgParser::getNonEmptyString(string prefix)
  throw (invalid_argument)
{
  string v = getString(prefix);
  if (v == "") {
    throw invalid_argument("value of option \"" + getOptName(prefix) + 
			   "\" is zero length string");
  }
  return v;
}

inline bool ArgParser::getBool(string prefix) throw (invalid_argument) {
  string optVal = getNonEmptyString(prefix);
  transform(optVal.begin(), optVal.end(), optVal.begin(),
	    (int (*)(int)) tolower);
  if (optVal == "true" || strtol(optVal.c_str(), 0, 10) != 0) {
    return true;
  } else {
    return false;
  }
}

inline bool ArgParser::getBool(string prefix, bool defaultVal)
  throw (invalid_argument)
{
  if (!hasOption(getOptName(prefix))) {
    return defaultVal;
  } else {
    return getBool(prefix);
  }
}

inline int ArgParser::getInt(string prefix) throw (invalid_argument) {
  string optVal = getNonEmptyString(prefix);
  char* p = 0;
  char** endp = &p;
  int i = strtol(optVal.c_str(), endp, 10);
  if (**endp != '\0') {
    throw invalid_argument("failed to convert the value \""+ optVal +
			   "\" of " + getOptName(prefix) + " to an int");
  }
  return i;
}

inline int ArgParser::getInt(string prefix, int defaultVal)
  throw (invalid_argument)
{
  if (!hasOption(getOptName(prefix))) {
    return defaultVal;
  } else {
    return getInt(prefix);
  }
}

inline float ArgParser::getFloat(string prefix) throw (invalid_argument) {
  string optVal = getNonEmptyString(prefix);
  char* p = 0;
  char** endp = &p;
  float f = strtof(optVal.c_str(), endp);
  if (**endp != '\0') {
    throw invalid_argument("failed to convert the value \""+ optVal +
			   "\" of " + getOptName(prefix) + " to a float");
  }
  return f;
}

inline float ArgParser::getFloat(string prefix, float defaultVal)
  throw (invalid_argument)
{
  if (!hasOption(getOptName(prefix))) {
    return defaultVal;
  } else {
    return getFloat(prefix);
  }
}

inline double ArgParser::getDouble(string prefix) throw (invalid_argument) {
  string optVal = getNonEmptyString(prefix);
  char* p = 0;
  char** endp = &p;
  double d = strtod(optVal.c_str(), endp);
  if (**endp != '\0') {
    throw invalid_argument("failed to convert the value \""+ optVal +
			   "\" of \"" + getOptName(prefix) + "\" to a double");
  }
  return d;
}

inline double ArgParser::getDouble(string prefix, double defaultVal)
  throw (invalid_argument)
{
  if (!hasOption(getOptName(prefix))) {
    return defaultVal;
  } else {
    return getDouble(prefix);
  }
}

inline string ArgParser::getOptName(string prefix) throw (invalid_argument) {
  if (prefix[0] == '-') {
    throw invalid_argument("option name cannot start with '-': " + prefix);
  }
  map<string, bool>::iterator pos = optSpecs.lower_bound(prefix);
  if (pos != optSpecs.end() &&
      pos->first.compare(0, prefix.size(), prefix) == 0)
  {  // found a match, need to make sure there is only one match
    string optName = pos->first;
    pos++;
    if (pos == optSpecs.end() ||
	pos->first.compare(0, prefix.size(), prefix) != 0)
    {
      return optName;
    } else {
      throw invalid_argument("more than one option spec matches prefix \"" +
			     prefix + '"');
    }
  } else {
    throw invalid_argument("failed to find an option spec corresponding " +
			   string("to prefix \"") + prefix + '"');
  }
}

#endif // _ARGPARSER_H_INCLUDED_
