#include <iostream>
#include <fstream>
#include "graph.h"
#include "priority-queue.h"


const int Node::INVALID_ID = INT_MAX;
const double Node::COST_MAX = INT_MAX;
const int Graph::MAX_N = INT_MAX;
const int Graph::MAX_L = INT_MAX;

const int Graph::PRECISION = 100;

Node::RoutTable* routingtablep = 0;
int srcNodeId = Node::INVALID_ID;
Graph* graphp = 0;

int comparator(int firstId, int secondId)
{
  double c1 = graphp->getCost(srcNodeId, firstId);
  double c2 = graphp->getCost(srcNodeId, secondId);

  return c1 < c2;
}

double Graph::getCost(int sId, int dId)
{
  const Node* src = nodes[sId];
  const Node* dst = nodes[dId];

  return src->getCostToDst(dst->getId(), *routingtablep);
}

Graph::Graph(const char* infile) {
	parseInput(infile);
	graphp = this; // used in comparator
	PQinit(getN(), comparator);
	compAllPairSP();
}


void Graph::parseInput(const char* infile) {
	ifstream inFileStream(infile, ios::in);

	if (!inFileStream || inFileStream.bad()) {
		cerr << "Failed to open input file: " <<  infile << endl;
	}

	int numNodes, numLinks;
  
	inFileStream >> numNodes >> numLinks;
  
	if (numNodes > 0 && numNodes < MAX_N) {
		_N = numNodes;
	} else {
		cerr << "Invalid number of graph nodes: " <<  numNodes << endl;
	}
	
	for (int i = 0; i < _N; i++) {
		Node* n = new Node(i, _N);

		nodes.push_back(n);
	}
	
	if (numLinks > 0 && numLinks < MAX_L) {
		_L = numLinks;
	} else {
		cerr << "Invalid number of graph links: " <<  numLinks;
	}

  /* We do not check for duplicate edges. The graph input file should
   * not contain duplicate links if it's a simple graph. Each link is
   * uni-directional.
   */
	
	for (int i = 0; i < numLinks; i++) {
		int nId, nbrId;
		double cost;
		
		inFileStream >> nId >> nbrId >> cost;
		//cout << nId << "  "  <<nbrId << " " << cost << endl ;
		map<int, double>& nbrs = nodes[nId]->nbrMap;

		
		if (nbrs.find(nbrId) == nbrs.end()) { 
		  // sanity check on input range is done in addNeighbor. 
		  nodes[nId]->addNeighbor(nodes[nbrId], cost);
		  nbrs.insert(make_pair(nbrId, cost));
		  Node::RoutTable* tab = new Node::RoutTable(_N);
		  nodes[nId]->delNbrTable.insert(make_pair(nbrId, tab));
		}
		
		Link alink(nId, nbrId);
		if (linkMap.find(alink) == linkMap.end()) {
			links.push_back(alink);
			linkMap.insert(make_pair(alink, i));
		} else {
			cerr << "duplicate links in input file: " << nId << ' ' << nbrId << endl;
			exit(1);
		}
	}

  // check for directed links

	for (map<Link, int>::const_iterator iter = linkMap.begin(); iter != linkMap.end(); ++iter) {
		Link const& curLink = iter->first;
		Link revLink(curLink.v, curLink.u);

		if (linkMap.find(revLink) == linkMap.end()) {
			cerr << "Input graph has directed link (" << curLink.u << ',' << curLink.v << ')' << endl;
			exit(1);
		}
	}
	inFileStream.close();
}


void Graph::compAllPairSP()
{
	nodeConnected = 0;
	for (vector<Node*>::iterator iter = nodes.begin(); iter != nodes.end(); ++iter) {
		Node* src = *iter;
		int srcId = src->getId();
    
		compOneSrcSP(src, src->routingTable);
    
		int connected = 0;

		for (int dst = 0; dst < _N; dst++) {
			if (src->routingTable[dst].nextHop != Node::INVALID_ID) {
				connected++;
			}
		}
		if (connected > nodeConnected) {
			nodeConnected = connected;
			repId = srcId;
		}
	}
	
	linkConnected = 0;
	for (vector<Node*>::iterator iter = nodes.begin(); iter < nodes.end(); ++iter) {
		Node * src = *iter;

		if (src->routingTable[repId].nextHop == Node::INVALID_ID) {
			continue;
		}
		linkConnected += src->neighbors.size();
	}
}

void Graph::compOneSrcSP(Node* src, Node::RoutTable& table) {
	srcNodeId = src->getId();  //used in comparator
	routingtablep = &table; // used in comparator

	for (int j = 0; j < getN(); j++) {
		PQinsert(j); // insert all nodeIds into the priority queue
	}

	while (!PQempty()) {
		// do relaxation around the top node in the queue
		const Node* minp = getNode(PQdelmin()); // next minimum cost node
		int minId = minp->getId();
		double cost = src->getCostToDst(minId, table); // the cost from i to np
      
		for (map<int, double>::const_iterator j = minp->nbrMap.begin(); j != minp->nbrMap.end(); ++j) {
			int jId = j->first;
	
			// note that distance between i and np must be known, not MY_INFINITY
			if (cost < Node::COST_MAX) { // np couldnot be unreachable from i
				double newCost = cost + j->second;
		  
				// if cost(src,min) + cost(min,j) < cost(src,j), set cost(src,j) to newcost
				if (newCost < src->getCostToDst(jId, table)) {
			
					/* updates src's routing table entry to j. updates cost,
					* src's next hop to j.
					*/
					src->updateEntry(jId, newCost, minId, table); 
					PQchange(jId);
				}
			}
		}
	}
}

void Graph::recordSPPerPair(Node* srcNp, int dst) {
	Node::Entry& e = srcNp->routingTable[dst];
	Path& shortest = e.shortest;
	
	int nextId = e.nextHop;
	Node* curNp = srcNp;

	// dst unreachable
	if (e.nextHop == Node::INVALID_ID) {
		disconnected++;
		return;
	}

	while (!hasReachDst(curNp,dst)) {
		shortest.nids.push_back(nextId);
		shortest.cost += curNp->nbrMap[nextId]; // adding cost
		curNp = nodes[nextId];
		nextId = curNp->routingTable[dst].nextHop;
	} 
  
}

void Graph::outputDegree(ostream& os) {
	
	map<int, int> pdf;
	for (vector<Node*>::const_iterator iter = nodes.begin(); iter != nodes.end(); ++iter) {
		Node* np = *iter;

		if (np->routingTable[repId].nextHop == Node::INVALID_ID) {
			continue;
		}
		int num = np->neighbors.size();
		map<int, int>::iterator found = pdf.find(num);
    
		if (found == pdf.end()) {
			pdf[num] = 1;
		} else {
			pdf[num]++;
		}
	}

	double sum = 0;
	double average = 0;
	int median = -1;
	int half = _N / 2;
	os << "# number | pdf | cdf | percentage" << endl ;
	for (map<int,int>::const_iterator iter = pdf.begin(); iter != pdf.end(); ++iter) {
		sum += iter->second;
		if (sum > half && median == -1) {
			median = iter->first;
		}
		average += iter->first * iter->second;
		os << iter->first << ' ' << iter->second << ' ' << sum << ' ' << (double) sum /_N << endl;
	}
	average /= sum;
	os << "# average Degree " << average << endl;
	os << "# median " << median << endl;
	os << "# total nodes " << nodes.size() << endl;
    os  << "# nodeConnected " << nodeConnected << endl;
    os  << "# linkConnected " << linkConnected << endl << endl;
}


void Path::print(ostream& os, vector<string> const* transtab)
{
	os.setf(ios_base::left, ios_base::adjustfield);
	for (vector<int>::iterator iter = nids.begin(); iter != nids.end(); ++iter) {
		if (transtab) {
			os  << (*transtab)[*iter] << ' ';
		} else {
			os  << *iter << ' ';
		}
	}
	os << endl;
}
