#include <vector>
#include <map>
#include <utility>
#include <limits.h>
#include <stdlib.h>

using namespace std;

struct Link {
	int u,v;
	Link() {}
	Link(int first, int second) : u(first), v(second) {}
		
	bool operator<(const Link& other) const {
		if (u != other.u) {
		  return u < other.u;
		} else {
		  return v < other.v;
		}
	}

	bool operator==(const Link& other) const {
		return u == other.u && v == other.v;
	}
	
};


// the path from this node to a destination
struct Path {
	Path():cost(0) {nids.clear();};
		
	void print(ostream& os, vector<string> const* tab);
	
	void clear() {
		cost = 0;
		nids.clear();
	}
	
	bool operator<(Path const& other) const {
		if (cost != other.cost) {
			return cost < other.cost;
		} else if (nids.size() != other.nids.size()) {
			return nids.size() < other.nids.size();
		} else {
			for (int i = 0; i < (int) nids.size(); i++) {
				if (nids[i] != other.nids[i]) {
					return nids[i] < other.nids[i];
				}
			}
		}
		return false;
	}
	  
	// nids does not include this node. nids.size() is the hop count.
	vector<int> nids;
	double cost;
};


struct Node {
    static const int INVALID_ID;
	static const double COST_MAX;
	
	  // A per destination entry.
	struct Entry {
		int dst;
		int nextHop; // next hop to dst
		int hopCount;
		double cost;

		Path shortest; // nodes (excluding src) that are on the shortest path

		Entry() {
			nextHop = INVALID_ID;
			hopCount = INT_MAX;
			cost = COST_MAX;
		}
		inline double getCost() const { return cost;}
  };


	struct Neighbor {
		Neighbor(Node* np, double c) {
			if (np == 0) {
				cerr << "Invalid neighbor pointer";
			}
			if (c < 0) {
				cerr << "Invalid neighbor cost";
			}
			nbr = np;
			cost = c;
		}
    
		int getId() const {
			return nbr->getId();
		}
		
		double getCostToNbr() const {
			return cost;
		}
		
		Node* nbr;
		double cost;
	};
	

	int id;	
	vector<Neighbor> neighbors;
	map<int, double> nbrMap;

	typedef vector<Entry> RoutTable;
	RoutTable routingTable;
	
	map<int, RoutTable*> delNbrTable;
	
	Node(int myid, int numNodes) { // n is the number of routing table entries
		id = myid;
		if (numNodes < 0) {
			cerr << "Invalid number of nodes in Node::init: " << numNodes << endl;
			exit(1);
		}
		routingTable.resize(numNodes);
		initSelfEntry(routingTable);
	}
	
	int getId() const {
		return id;
	}
	
	void addNeighbor(Node* np, double c) {
		neighbors.push_back(Neighbor(np,c));

		int nbrId = np->getId();
		Entry& e = routingTable[nbrId];

		e.hopCount = 1;
		e.cost = c;
		e.nextHop = nbrId;
		e.dst = nbrId;  
	}
	
	void updateEntry(int dstId, double newCost, int prevId, RoutTable& tab) {
		Entry& dstEntry = tab[dstId];
		const Entry& prevEntry = tab[prevId];

		dstEntry.cost = newCost;
		dstEntry.nextHop = prevEntry.nextHop;
		dstEntry.hopCount = prevEntry.hopCount + 1;
		dstEntry.dst = dstId;
	}
	
	void printNeighbors() {
		for (vector<Neighbor>::iterator iter = neighbors.begin(); iter != neighbors.end(); ++iter) {
			cout << iter->getId() << " with cost  "<< iter->getCostToNbr()  << endl ;
		}
	}
	
	void printRoutingTable() {

		for (vector<Entry>::iterator iter = routingTable.begin(); iter != routingTable.end(); ++iter) {
			cout << "To reach " << iter->dst << " cost  " << iter->cost
				<< " using next Hop " << iter->nextHop  << 
				" in " << iter->hopCount << " hops" << endl ;
		}
	}
	
	double getCostToDst(int dstId, RoutTable& table) const {
		if (dstId < 0 || dstId > (int) table.size()) {
			cerr << "dstId out of range in Node::getCostToDst: " << endl;
		}
		return table[dstId].cost;
	}	
	
	double getCostToDst(int dstId) const {
		if (dstId < 0 || dstId > (int) routingTable.size()) {
			cerr << "dstId out of range in Node::getCostToDst: " << endl;
		}
		return routingTable[dstId].cost;
	}
	
	int getNextHopToDst(int dstId) const {
		if (dstId < 0 || dstId > (int) routingTable.size()) {
			cerr << "dstId out of range in Node::getCostToDst: " << endl;
		}
		return routingTable[dstId].nextHop;
	}
	
	vector<Node *> getNeighbors() {
		vector<Node *> nbrsList;
		
		for (vector<Neighbor>::iterator iter = neighbors.begin(); iter != neighbors.end(); ++iter) {
			nbrsList.push_back(iter->nbr);
		}
		return nbrsList;
	}
	
	inline void initSelfEntry(RoutTable& tab) {
		Entry& e = tab[id];
	
		e.cost = 0;
		e.nextHop = id;
		e.hopCount = 0;
		e.dst = id;
	}
	
};

struct Graph {
	static const int MAX_N;
	static const int MAX_L;
	static const int PRECISION;
	int _N;
	int _L;
	
	vector<Node*> nodes;
	vector<Link> links;
	map<Link, int> linkMap;
	int repId; // a node Id from the maximally connected component
	int nodeConnected; // number of nodes in the maximally connected component
	int linkConnected; // number of links in the maximally connected component
	int disconnected;
	
	map<int, int> pathPdf;

	
	Graph(const char* infile);
	void parseInput(const char*);
	void compAllPairSP(); // shortest path computation
	void compOneSrcSP(Node* node, Node::RoutTable& table);
	void recordSPPerPair(Node* srcNp, int dst);
	double getCost(int sId, int dId);
	
	int const getN() { return _N ; } 
	const Node* const getNode(int nId) { return nodes[nId];}
	
	bool hasReachDst(Node* srcNp, int dst) const {
		return srcNp->routingTable[dst].nextHop == srcNp->getId();
	}
	
	
	double getLinkCost(int srcId, int dstId) {
		return nodes[srcId]->nbrMap[dstId];
	}

	void outputDegree(ostream& os);
};
